/* Copyright 2019 SiFive, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You should have received a copy of LICENSE.Apache2 along with
 * this software. If not, you may obtain a copy at
 *
 *    https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

body {
  hyphens: auto;
  background-color: black;
  font-family: 'Roboto', sans-serif;
}

th,td {
  padding: 2px 8px;
  border: 1px solid #000;
}

div { color: white; }

pre {
  /* font-family: 'IBM Plex Mono', monospace; */
  font-family: 'Roboto Mono', monospace;
}

.Program    { color: #999; }
.VarRef     { color: white; }
.VarDef     { color: #ee3; }
.VarArg     { color: #55f; }
.Prim       { color: yellow; }
.Literal    { color: pink; }

*[sourceType]:hover { background-color: hsla(0, 100%, 100%, 0.1); }
*[focus='true']:hover { background-color: green; }

.tooltip {
  display: block;
  padding: 10px;
  z-index: 5;
  position: absolute;
  min-width: 200px;
  background-color: #333;
}

a:link {
  text-decoration: none;
}
